﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Forms;
using OpcLabs.BaseLib.Collections.Generic.Extensions;
using OpcLabs.BaseLib.Extensions;
using OpcLabs.BaseLib.Extensions.Internal;
using OpcLabs.BaseLib.Forms.Common;
using OpcLabs.BaseLib.Forms.IdentityModel;
using OpcLabs.BaseLib.Forms.Testing;
using OpcLabs.BaseLib.LiveBinding;
using OpcLabs.BaseLib.LiveMapping;
using OpcLabs.BaseLib.Navigation;
using OpcLabs.BaseLib.OperationModel;
using OpcLabs.BaseLib.Persistence;

namespace OpcLabs.BaseLibFormsTestApp
{
    static class Program
    {
        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        [STAThread]
        static void Main()
        {
#if (NETSTANDARD)
            Application.SetHighDpiMode(HighDpiMode.SystemAware);
#endif
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);

            var switchboardForm = new SwitchboardForm
            {
                Text = "BaseLibForms Test Application",
                CommandDictionary =
                {
                    ["ObjectTestbed"] = ObjectTestbed,
                    ["MemberNameEditor"] = MemberNameEditor,
                    ["ArgumentsMemberPathEditor"] = ArgumentsMemberPathEditor,
                    ["TimeSpanInput"] = TimeSpanInput,
                    ["BrowseNodesTestSample"] = BrowseNodesTestSample,
                    ["SingleExceptionForm"] = SingleExceptionForm,
                    ["SingleExceptionForm_Bug1"] = SingleExceptionForm_Bug1,
                    ["MultiExceptionForm"] = MultiExceptionForm,
                    ["NotifierTestForm"] = NotifierTestForm,
                    ["NumericUpDown"] = HexadecimalNumericUpDown,
                    ["DateTimeTestForm"] = DateTimeTestForm,
                    ["TraverserTestForm"] = TraverserTestForm,
                    ["Design"] = DesignForm,
                    ["Dialogs"] = Dialogs
                }
            };

            Application.Run(switchboardForm);
        }

        static private void ObjectTestbed()
        {
            var objectTypes = new List<Type>();
            objectTypes.AddRange(
                typeof(ArrayExtension).Assembly.GetTypes()
                .Concat(typeof(HandleArguments).Assembly.GetTypes())
                .Concat(typeof(MemberInput).Assembly.GetTypes())
                .Concat(typeof(Program).Assembly.GetTypes())
                );
            var favoriteObjectTypes = new List<Type>
                {
                    typeof (BrowsePath),
                };
            Type primaryType = typeof(BrowsePath);

            using (var form = new ObjectTestbedForm())
            {
                form.SetOnlyUsableTypes(objectTypes);
                form.FavoriteTypes = favoriteObjectTypes;
                form.CurrentType = primaryType;
                var deferredInstanceInjector = new DeferredInstanceInjector(form.Instances);
                // ... 
                form.Instances = deferredInstanceInjector;
                form.InitialAction = ObjectTestbedFormAction.Instantiate;
                form.ImportedNamespaces = new[] { "OpcLabs.BaseLib.Diagnostics", "OpcLabs.BaseLib.Extensions" };

                form.ShowDialog();
            }
        }

        static private void DesignForm()
        {
            // TODO: NETCORE-REVIEW:
#if (NETFRAMEWORK)
            using (var form = new Design.Form1())
            {
                form.ShowDialog();
            }
#endif
        }

        static internal void Dialogs()
        {
            var dialogTypes = DialogsForm.GetConcreteDialogTypes(typeof(UserIdentityDialog).Assembly);
            var form = new DialogsForm();
            form.Types.AddRange(dialogTypes);
            form.Show();
        }

        static private void MemberNameEditor()
        {
            using (var form = new MemberNameEditorTestForm())
            {
                form.ShowDialog();
            }
        }
        static private void HexadecimalNumericUpDown()
        {
            using (var form = new HexadecimalNumericUpDownTestForm())
            {
                form.ShowDialog();
            }
        }

        static private void DateTimeTestForm()
        {
            using (var form = new DateTimeTestForm())
            {
                form.ShowDialog();
            }
        }

        static private void TraverserTestForm()
        {
            using (var form = new TraverserTestForm())
            {
                form.ShowDialog();
            }
        }

        static private void ArgumentsMemberPathEditor()
        {
            var pointBinding = new PointBinding(new Form(), new Form(), new Form(), new Form());
            using (var form = new PropertyGridForm {GridObject = pointBinding})
            {
                form.ShowDialog();
            }
        }

        static private void NotifierTestForm()
        {
            using (var form = new NotifierTestForm())
            {
                form.ShowDialog();
            }
        }

        static private void TimeSpanInput()
        {
            using (var form = new TimeSpanInputTestForm())
            {
                form.ShowDialog();
            }
        }

        static private void BrowseNodesTestSample()
        {
            using (var form = new BrowseNodesTestSample())
            {
                form.ShowDialog();
            }
        }

        static private void SingleExceptionForm()
        {
            try
            {
// ReSharper disable ConvertToConstant.Local
                var n = 0;
// ReSharper restore ConvertToConstant.Local
#pragma warning disable 168
                // ReSharper disable once UnusedVariable
                var i = 5 / n;
#pragma warning restore 168
            }
            catch (Exception exception)
            {
                ExceptionBox.ShowOne(null, exception);
            }
        }

        // ReSharper disable InconsistentNaming
        static private void SingleExceptionForm_Bug1()
        // ReSharper restore InconsistentNaming
        {
            //var exception = new Exception("1\n2\n3\n4", new Exception("1\n2\n3\n4", new Exception("1\n2\n3\n4")));
            var exception = new Exception("Problem browsing OPC Unified Architecture nodes.",
                new Exception("An OPC-UA operation failure with error code -2138963968 (0x80820000) occurred, originating from 'Opc.Ua.Core'. The inner OPC-UA service exception with service result 'BadTcpInternalError' contains details about the problem.", 
                    new Exception(@"	Opc.Ua.Core	OPC-UA service result - Error establishing a connection.
---- SERVICE RESULT ----
StatusCode: {BadTcpInternalError} = 0x80820000 (2156003328)
Description: Error establishing a connection.
-----
StatusCode: {BadTcpInternalError} = 0x80820000 (2156003328)
Description: An unexpected error occurred while connecting to the server.
AdditionalInfo: >>> An unexpected error occurred while connecting to the server.
---    at Opc.Ua.Bindings.TcpAsyncOperation`1.End(Int32 timeout)
---    at Opc.Ua.Bindings.TcpClientChannel.OnConnectOnDemandComplete(Object state)
-----
StatusCode: {Bad} = 0x80000000 (2147483648)
Description: No connection could be made because the target machine actively refused it 127.0.0.1:1111
AdditionalInfo: >>> No connection could be made because the target machine actively refused it 127.0.0.1:1111
---    at System.Net.Sockets.Socket.EndConnect(IAsyncResult asyncResult)
---    at Opc.Ua.Bindings.TcpMessageSocket.EndConnect(IAsyncResult result)
---    at Opc.Ua.Bindings.TcpClientChannel.OnConnectComplete(IAsyncResult result)
---- REMARKS ----
Some possible causes of this error are that the OPC-UA server is not running, or is not configured to listen on the specified port. Also, the network connection may be broken (cable unplugged?).")));
            ExceptionBox.ShowOne(null, exception);
        }

        static private void MultiExceptionForm()
        {
            var exceptions = new List<Exception>();
            try
            {
// ReSharper disable ConvertToConstant.Local
                var n = 0;
// ReSharper restore ConvertToConstant.Local
#pragma warning disable 168
                // ReSharper disable once UnusedVariable
                var i = 5 / n;
#pragma warning restore 168
            }
            catch (Exception exception)
            {
                exceptions.Add(exception);
            }

            var exception2 = new InvalidOperationException("Invalid operation",
                                                          new ExecutionException("Execution exception"))
                {
                    HelpLink = "Help link: www.opclabs.com"
                };
            exceptions.Add(exception2);

            try
            {
// ReSharper disable ConvertToConstant.Local
                var n = 0;
// ReSharper restore ConvertToConstant.Local
#pragma warning disable 168
                // ReSharper disable once UnusedVariable
                var i = 10 / n;
#pragma warning restore 168
            }
            catch (Exception exception)
            {
                exceptions.Add(exception);
            }
            ExceptionBox.ShowMultiple(null, exceptions);
        }
    }
}
