﻿using System;
using System.IO;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.Extensions.FileProviders;
using Microsoft.Extensions.Primitives;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Diagnostics;
using OpcLabs.BaseLib.Extensions.FileProviders;
using OpcLabs.UnitTestLib;

namespace OpcLabs.BaseLibTest.Extensions.FileProviders
{
    // ReSharper disable once InconsistentNaming
    public abstract class IWritableFileProviderTest
    {
        #region Static Private Fields

        [NotNull]
        static private readonly Random Random = new Random(Environment.TickCount);

        #endregion



        #region Test Methods

        [TestMethod]
        public void CreateFailingFileProviderTest()
        {
            CreateFileProvider(failing: true);
        }

        [TestMethod]
        public void CreateFailingFileProvider2Test()
        {
            CreateFileProvider2(failing: true);
        }

        [TestMethod]
        public void CreateFailingWritableFileProviderTest()
        {
            CreateWritableFileProvider(failing: true);
        }

        [TestMethod]
        public void CreateFileProviderTest()
        {
            CreateFileProvider(failing:false);
        }

        [TestMethod]
        public void CreateFileProvider2Test()
        {
            CreateFileProvider2(failing: false);
        }

        [TestMethod]
        public void CreateWritableFileProviderTest()
        {
            CreateWritableFileProvider(failing: false);
        }

        // IDirectoryContents Tests

        [TestMethod]
        public void IDirectoryContents_CompoundPath_WithAFile_Enumeration_Test() =>
            IDirectoryContents_CompoundPath_WithAFile_Enumeration_Test(path => GetDirectoryContents(path));

        [TestMethod]
        public void IDirectoryContents_EmptyPath_Enumeration_Test() => IDirectoryContents_EmptyPath_Enumeration_Test(
            path => GetDirectoryContents(path));

        [TestMethod]
        public void IDirectoryContents_EmptyPath_ToString_Test() => 
            IDirectoryContents_EmptyPath_ToString_Test(path => GetDirectoryContents(path));

        [TestMethod]
        public void IDirectoryContents_EmptyPath_WithAFile_Enumeration_Test() =>
            IDirectoryContents_EmptyPath_WithAFile_Enumeration_Test(path => GetDirectoryContents(path));

        [TestMethod]
        public void IDirectoryContents_Exists_Negative_Test() =>
            IDirectoryContents_Exists_Negative_Test(path => GetDirectoryContents(path));

        [TestMethod]
        public void IDirectoryContents_NullPath_Enumeration_Test() =>
            IDirectoryContents_NullPath_Enumeration_Test(path => GetDirectoryContents(path));

        [TestMethod]
        public void IDirectoryContents_SimplePath_WithAFile_Enumeration_Test() =>
            IDirectoryContents_SimplePath_WithAFile_Enumeration_Test(path => GetDirectoryContents(path));

        // IDirectoryContents2 Tests

        [TestMethod]
        public void IDirectoryContents2_CompoundPath_WithAFile_Enumeration_Test() =>
            IDirectoryContents_CompoundPath_WithAFile_Enumeration_Test(path => GetDirectoryContents2(path));

        [TestMethod]
        public void IDirectoryContents2_EmptyPath_Enumeration_Test() => IDirectoryContents_EmptyPath_Enumeration_Test(
            path => GetDirectoryContents2(path));

        [TestMethod]
        public void IDirectoryContents2_EmptyPath_ToString_Test() =>
            IDirectoryContents_EmptyPath_ToString_Test(path => GetDirectoryContents2(path));

        [TestMethod]
        public void IDirectoryContents2_EmptyPath_WithAFile_Enumeration_Test() =>
            IDirectoryContents_EmptyPath_WithAFile_Enumeration_Test(path => GetDirectoryContents2(path));

        [TestMethod]
        public void IDirectoryContents2_Exists_Negative_Test() =>
            IDirectoryContents_Exists_Negative_Test(path => GetDirectoryContents2(path));

        [TestMethod]
        public void IDirectoryContents2_NullPath_Enumeration_Test() =>
            IDirectoryContents_NullPath_Enumeration_Test(path => GetDirectoryContents2(path));

        [TestMethod]
        public void IDirectoryContents_GetStatistics_NonRecursive_Test() => IDirectoryContents_GetStatistics_Test(recursive: false);

        [TestMethod]
        public void IDirectoryContents_GetStatistics_Recursive_Test() => IDirectoryContents_GetStatistics_Test(recursive:true);

        [TestMethod]
        public void IDirectoryContents2_SimplePath_WithAFile_Enumeration_Test() =>
            IDirectoryContents_SimplePath_WithAFile_Enumeration_Test(path => GetDirectoryContents2(path));

        // IFileInfo Tests

        [TestMethod]
        public void IFileInfo_CreateReadStream_CompoundPath_Test() =>
            IFileInfo_CreateReadStream_CompoundPath_Test(path => GetFileInfo(path));

        [ExpectedException(typeof(FileNotFoundException))]
        [TestMethod]
        public void IFileInfo_CreateReadStream_NonExistent_Test() =>
            IFileInfo_CreateReadStream_NonExistent_Test(path => GetFileInfo(path));

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Empty_0B_Test() =>
            IFileInfo_CreateReadStream_Read_Empty_Test(path => GetFileInfo(path), 0);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Empty_1B_Test() =>
            IFileInfo_CreateReadStream_Read_Empty_Test(path => GetFileInfo(path), 1);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Empty_16B_Test() =>
            IFileInfo_CreateReadStream_Read_Empty_Test(path => GetFileInfo(path), 16);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Empty_256B_Test() =>
            IFileInfo_CreateReadStream_Read_Empty_Test(path => GetFileInfo(path), 256);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Empty_4KiB_Test() =>
            IFileInfo_CreateReadStream_Read_Empty_Test(path => GetFileInfo(path), 4 * 1024);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Empty_8KiB_Test() =>
            IFileInfo_CreateReadStream_Read_Empty_Test(path => GetFileInfo(path), 8 * 1024);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Empty_8KiB_Plus_1_Test() =>
            IFileInfo_CreateReadStream_Read_Empty_Test(path => GetFileInfo(path), 8 * 1024 + 1);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Empty_16KiB_Test() =>
            IFileInfo_CreateReadStream_Read_Empty_Test(path => GetFileInfo(path), 16 * 1024);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Empty_32KiB_Test() =>
            IFileInfo_CreateReadStream_Read_Empty_Test(path => GetFileInfo(path), 32 * 1024);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Empty_64KiB_Test() =>
            IFileInfo_CreateReadStream_Read_Empty_Test(path => GetFileInfo(path), 64 * 1024);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Empty_128KiB_Test() =>
            IFileInfo_CreateReadStream_Read_Empty_Test(path => GetFileInfo(path), 128 * 1024);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Full_0B_Test() =>
            IFileInfo_CreateReadStream_Read_Full_Test(path => GetFileInfo(path), 0);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Full_1B_Test() =>
            IFileInfo_CreateReadStream_Read_Full_Test(path => GetFileInfo(path), 1);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Full_16B_Test() =>
            IFileInfo_CreateReadStream_Read_Full_Test(path => GetFileInfo(path), 16);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Full_256B_Test() =>
            IFileInfo_CreateReadStream_Read_Full_Test(path => GetFileInfo(path), 256);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Full_4KiB_Test() =>
            IFileInfo_CreateReadStream_Read_Full_Test(path => GetFileInfo(path), 4 * 1024);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Full_8KiB_Test() =>
            IFileInfo_CreateReadStream_Read_Full_Test(path => GetFileInfo(path), 8 * 1024);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Full_16KiB_Test() =>
            IFileInfo_CreateReadStream_Read_Full_Test(path => GetFileInfo(path), 16 * 1024);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Full_32KiB_Test() =>
            IFileInfo_CreateReadStream_Read_Full_Test(path => GetFileInfo(path), 32 * 1024);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Full_64KiB_Test() =>
            IFileInfo_CreateReadStream_Read_Full_Test(path => GetFileInfo(path), 64 * 1024);

        [TestMethod]
        public void IFileInfo_CreateReadStream_Read_Full_128KiB_Test() =>
            IFileInfo_CreateReadStream_Read_Full_Test(path => GetFileInfo(path), 128 * 1024);

        [TestMethod]
        public void IFileInfo_CreateReadStream_SimplePath_Test() => 
            IFileInfo_CreateReadStream_SimplePath_Test(path => GetFileInfo(path));

        [TestMethod]
        public void IFileInfo_Exists_Negative_Test() => IFileInfo_Exists_Negative_Test(path => GetFileInfo(path));

        [TestMethod]
        public void IFileInfo_IsDirectory_NonExistent_Test() => 
            IFileInfo_IsDirectory_NonExistent_Test(path => GetFileInfo(path));

        [TestMethod]
        public void IFileInfo_Length_DeletedFile_Test() => IFileInfo_Length_DeletedFile_Test(path => GetFileInfo(path));

        [TestMethod]
        public void IFileInfo_Length_NonExistent_Test() => IFileInfo_Length_NonExistent_Test(path => GetFileInfo(path));

        [TestMethod]
        public void IFileInfo_Name_Empty_Test() => IFileInfo_Name_Empty_Test(path => GetFileInfo(path));

        [TestMethod]
        public void IFileInfo_Name_Null_Test() => IFileInfo_Name_Null_Test(path => GetFileInfo(path));

        // IFileInfo2 Tests

        [TestMethod]
        public void IFileInfo2_CreateReadStream_CompoundPath_Test() =>
            IFileInfo_CreateReadStream_CompoundPath_Test(path => GetFileInfo2(path));

        [ExpectedException(typeof(FileNotFoundException))]
        [TestMethod]
        public void IFileInfo2_CreateReadStream_NonExistent_Test() =>
            IFileInfo_CreateReadStream_NonExistent_Test(path => GetFileInfo2(path));

        [TestMethod]
        public void IFileInfo2_CreateReadStream_Read_Empty_0B_Test() =>
            IFileInfo_CreateReadStream_Read_Empty_Test(path => GetFileInfo2(path), 0);

        [TestMethod]
        public void IFileInfo2_CreateReadStream_Read_Empty_1B_Test() =>
            IFileInfo_CreateReadStream_Read_Empty_Test(path => GetFileInfo2(path), 1);

        [TestMethod]
        public void IFileInfo2_CreateReadStream_Read_Full_0B_Test() =>
            IFileInfo_CreateReadStream_Read_Full_Test(path => GetFileInfo2(path), 0);

        [TestMethod]
        public void IFileInfo2_CreateReadStream_Read_Full_1B_Test() =>
            IFileInfo_CreateReadStream_Read_Full_Test(path => GetFileInfo2(path), 1);
        
        [TestMethod]
        public void IFileInfo2_CreateReadStream_SimplePath_Test() =>
            IFileInfo_CreateReadStream_SimplePath_Test(path => GetFileInfo2(path));

        [TestMethod]
        public void IFileInfo2_Exists_Negative_Test() => IFileInfo_Exists_Negative_Test(path => GetFileInfo2(path));

        [TestMethod]
        public void IFileInfo2_IsDirectory_NonExistent_Test() =>
            IFileInfo_IsDirectory_NonExistent_Test(path => GetFileInfo2(path));

        [TestMethod]
        public void IFileInfo2_Length_DeletedFile_Test() => IFileInfo_Length_DeletedFile_Test(path => GetFileInfo2(path));

        [TestMethod]
        public void IFileInfo2_Length_NonExistent_Test() => IFileInfo_Length_NonExistent_Test(path => GetFileInfo2(path));

        [TestMethod]
        public void IFileInfo2_Name_Empty_Test() => IFileInfo_Name_Empty_Test(path => GetFileInfo2(path));

        [TestMethod]
        public void IFileInfo2_Name_Null_Test() => IFileInfo_Name_Null_Test(path => GetFileInfo2(path));

        // IFileProvider Tests

        [ExpectedException(typeof(IOException))]
        [TestMethod]
        public void IFileProvider_GetDirectoryContents_Failure_Test()
        {
            string filePath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IDirectoryContents directoryContents = GetDirectoryContents(filePath, failing: true);
            // ReSharper disable once ReturnValueOfPureMethodIsNotUsed
            directoryContents.GetEnumerator();  // initiate the enumeration
        }

        [TestMethod]
        public void IFileProvider_GetDirectoryContents_NonExistent_Test()
        {
            string filePath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IDirectoryContents directoryContents = GetDirectoryContents(filePath, failing: false);
            Assert.IsFalse(directoryContents.Exists);
            Assert.IsFalse(directoryContents.Any());
        }

        [TestMethod]
        public void IFileProvider_GetFileInfo_Empty_Test()
        {
            GetFileInfo("", failing: true);
        }

        [TestMethod]
        public void IFileProvider_GetFileInfo_Empty_Test2()
        {
            GetFileInfo("", failing: false);
        }

        [TestMethod]
        public void IFileProvider_GetFileInfo_Empty_Then_NotExists_Test()
        {
            IFileInfo fileInfo = GetFileInfo("", failing: true);
            Assert.IsFalse(fileInfo.Exists);
        }


        [TestMethod]
        public void IFileProvider_GetFileInfo_Empty_Then_NotExists_Test2()
        {
            IFileInfo fileInfo = GetFileInfo("", failing: false);
            Assert.IsFalse(fileInfo.Exists);
        }

        [ExpectedException(typeof(IOException))]
        [TestMethod]
        public void IFileProvider_GetFileInfo_Failure_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IFileInfo fileInfo = GetFileInfo(filePath, failing: true);
            long _ = fileInfo.Length;   // trigger the error
        }

        [TestMethod]
        public void IFileProvider_GetFileInfo_Null_Test()
        {
            GetFileInfo(null, failing: true);
        }

        [TestMethod]
        public void IFileProvider_GetFileInfo_Null_Test2()
        {
            GetFileInfo(null, failing: false);
        }

        [TestMethod]
        public void IFileProvider_GetFileInfo_Null_Then_NotExists_Test()
        {
            IFileInfo fileInfo = GetFileInfo(null, failing: true);
            Assert.IsFalse(fileInfo.Exists);
        }

        [TestMethod]
        public void IFileProvider_GetFileInfo_Null_Then_NotExists_Test2()
        {
            IFileInfo fileInfo = GetFileInfo(null, failing: false);
            Assert.IsFalse(fileInfo.Exists);
        }

        [TestMethod]
        public void IFileProvider_ToString_Test() => IFileProvider_ToString_Test(CreateFileProvider(failing: false));

        [TestMethod]
        public void IFileProvider_Watch_Null_Test() => IFileProvider_Watch_Null_Test(CreateFileProvider(failing: false));

        // IFileProvider2 Tests

        [TestMethod]
        public void IFileProvider2_ToString_Test() => IFileProvider_ToString_Test(CreateFileProvider2(failing: false));

        [TestMethod]
        public void IFileProvider2_Watch_Null_Test() => IFileProvider_Watch_Null_Test(CreateFileProvider2(failing: false));

        // IWritableDirectoryContents Tests

        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void IWritableDirectoryContents_CompoundPath_CreateSubdirectory_CompoundPath_Test()
        {
            string directoryPath0 = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath0);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath0);
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            writableDirectoryContents.CreateSubdirectory(directoryPath);
        }

        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void IWritableDirectoryContents_CompoundPath_CreateSubdirectory_CompoundPath_Twice_Test()
        {
            string directoryPath0 = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath0);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath0);
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            writableDirectoryContents.CreateSubdirectory(directoryPath);
            writableDirectoryContents.CreateSubdirectory(directoryPath);
        }

        [ExpectedException(typeof(ArgumentNullException))]
        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void IWritableDirectoryContents_CompoundPath_CreateSubdirectory_Null_Test()
        {
            string directoryPath0 = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath0);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath0);
            // ReSharper disable once AssignNullToNotNullAttribute
            writableDirectoryContents.CreateSubdirectory(null);
        }

        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void IWritableDirectoryContents_CompoundPath_CreateSubdirectory_SimplePath_Test()
        {
            string directoryPath0 = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath0);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath0);
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            writableDirectoryContents.CreateSubdirectory(directoryPath);
        }

        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void IWritableDirectoryContents_CompoundPath_CreateSubdirectory_SimplePath_Twice_Test()
        {
            string directoryPath0 = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath0);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath0);
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            writableDirectoryContents.CreateSubdirectory(directoryPath);
            writableDirectoryContents.CreateSubdirectory(directoryPath);
        }

        [TestMethod]
        public void IWritableDirectoryContents_CompoundPath_Create_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.Create();
        }

        [TestMethod]
        public void IWritableDirectoryContents_CompoundPath_Create_Twice_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.Create();
            writableDirectoryContents.Create();
        }

        [TestMethod]
        public void IWritableDirectoryContents_CompoundPath_Create_Then_Exists_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.Create();
            Assert.IsTrue(writableDirectoryContents.Exists);
        }

        [ExpectedException(typeof(IOException))]
        [TestMethod]
        public void IWritableDirectoryContents_CompoundPath_Delete_NonRecursive_NotEmpty_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.Create();
            string filePath = FormattableString.Invariant($"{directoryPath}\\TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            CreateAndDisposeOfWriteStream(filePath);
            writableDirectoryContents.Delete(recursive: false);
        }

        [TestMethod]
        public void IWritableDirectoryContents_CompoundPath_Delete_NonRecursive_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.Create();
            writableDirectoryContents.Delete(recursive: false);
        }

        [TestMethod]
        public void IWritableDirectoryContents_CompoundPath_Delete_Recursive_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.Create();
            writableDirectoryContents.Delete(recursive: true);
        }

        [TestMethod]
        public void IWritableDirectoryContents_CompoundPath_WithAFile_Enumeration_Test() =>
            IWritableDirectoryContents_CompoundPath_WithAFile_Enumeration_Test(path => GetWritableDirectoryContents(path));

        [TestMethod]
        public void IWritableDirectoryContents_EmptyPath_Enumeration_Test() => IWritableDirectoryContents_EmptyPath_Enumeration_Test(
            path => GetWritableDirectoryContents(path));

        [TestMethod]
        public void IWritableDirectoryContents_EmptyPath_WithAFile_Enumeration_Test() =>
            IWritableDirectoryContents_EmptyPath_WithAFile_Enumeration_Test(path => GetWritableDirectoryContents(path));

        [TestMethod]
        public void IWritableDirectoryContents_NullPath_Enumeration_Test() =>
            IWritableDirectoryContents_NullPath_Enumeration_Test(path => GetWritableDirectoryContents(path));

        [TestMethod]
        public void IWritableDirectoryContents_Root_Create_Test()
        {
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(null);
            writableDirectoryContents.Create(); // must not throw
        }

        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void IWritableDirectoryContents_Root_CreateSubdirectory_CompoundPath_Test()
        {
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(null);
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            writableDirectoryContents.CreateSubdirectory(directoryPath);
        }

        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void IWritableDirectoryContents_Root_CreateSubdirectory_CompoundPath_Twice_Test()
        {
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(null);
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            writableDirectoryContents.CreateSubdirectory(directoryPath);
            writableDirectoryContents.CreateSubdirectory(directoryPath);
        }

        [ExpectedException(typeof(IOException))]
        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void IWritableDirectoryContents_Root_CreateSubdirectory_Invalid_Test()
        {
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(null);
            // ReSharper disable once AssignNullToNotNullAttribute
            writableDirectoryContents.CreateSubdirectory(new string(Path.GetInvalidPathChars()));
        }

        [ExpectedException(typeof(ArgumentNullException))]
        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void IWritableDirectoryContents_Root_CreateSubdirectory_Null_Test()
        {
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(null);
            // ReSharper disable once AssignNullToNotNullAttribute
            writableDirectoryContents.CreateSubdirectory(null);
        }

        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void IWritableDirectoryContents_Root_CreateSubdirectory_SimplePath_Test()
        {
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(null);
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            writableDirectoryContents.CreateSubdirectory(directoryPath);
        }

        [ExpectedException(typeof(IOException))]
        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void IWritableDirectoryContents_Root_CreateSubdirectory_SimplePath_Failure_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            CreateAndDisposeOfWriteStream(filePath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(null);
            writableDirectoryContents.CreateSubdirectory(filePath);
        }

        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void IWritableDirectoryContents_Root_CreateSubdirectory_SimplePath_Twice_Test()
        {
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(null);
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            writableDirectoryContents.CreateSubdirectory(directoryPath);
            writableDirectoryContents.CreateSubdirectory(directoryPath);
        }

        [ExpectedException(typeof(IOException))]
        [TestMethod]
        public void IWritableDirectoryContents_Root_Delete_NonRecursive_Test()
        {
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(null);
            writableDirectoryContents.Delete();
        }

        [ExpectedException(typeof(UnauthorizedAccessException))]
        [TestMethod]
        public void IWritableDirectoryContents_Root_Delete_Recursive_Test()
        {
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(null);
            writableDirectoryContents.Delete(recursive:true);
        }

        [ExpectedException(typeof(IOException))]
        [TestMethod]
        public void IWritableDirectoryContents_Root_MoveTo_Test()
        {
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(null);
            string destinationDirectoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            writableDirectoryContents.MoveTo(destinationDirectoryPath);
        }

        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_Create_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.Create();
        }

        [ExpectedException(typeof(IOException))]
        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_Create_Failure_Test()
        {
            string filePath = FormattableString.Invariant($"TestPath-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            CreateAndDisposeOfWriteStream(filePath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(filePath);
            writableDirectoryContents.Create();
        }

        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_Create_Twice_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.Create();
            writableDirectoryContents.Create();
        }

        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_Create_Then_Exists_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.Create();
            Assert.IsTrue(writableDirectoryContents.Exists);
        }

        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_Delete_NonRecursive_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.Create();
            writableDirectoryContents.Delete(recursive: false);
        }

        [ExpectedException(typeof(IOException))]
        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_Delete_NonRecursive_NotEmpty_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.Create();
            string filePath = FormattableString.Invariant($"{directoryPath}\\TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            CreateAndDisposeOfWriteStream(filePath);
            writableDirectoryContents.Delete(recursive: false);
        }

        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_Delete_Recursive_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.Create();
            writableDirectoryContents.Delete(recursive:true);
        }

        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_Exist_Failure_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath, failing:true);
            Assert.IsFalse(writableDirectoryContents.Exists);
        }

        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_Exist_Negative_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            Assert.IsFalse(writableDirectoryContents.Exists);
        }

        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_MoveTo_CompoundPath_Absolute_Test()
        {
            string directoryPath1 = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath1);
            IWritableDirectoryContents writableDirectoryContents1 = GetWritableDirectoryContents(directoryPath1);
            writableDirectoryContents1.Create();

            string directoryPath2 = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath2);
            IWritableDirectoryContents writableDirectoryContents2 = GetWritableDirectoryContents(directoryPath2);
            writableDirectoryContents2.Create();

            string destinationDirectoryPath = FormattableString.Invariant($"\\{directoryPath2}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(destinationDirectoryPath);
            writableDirectoryContents1.MoveTo(destinationDirectoryPath);
        }

        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_MoveTo_CompoundPath_RelativeOutside_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.Create();
            string destinationDirectoryPath = FormattableString.Invariant($"../TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(destinationDirectoryPath);
            writableDirectoryContents.MoveTo(destinationDirectoryPath);
        }

        [ExpectedException(typeof(DirectoryNotFoundException))]
        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_MoveTo_CompoundPath_RelativeNotFound_Test()
        {
            string directoryPath1 = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath1);
            IWritableDirectoryContents writableDirectoryContents1 = GetWritableDirectoryContents(directoryPath1);
            writableDirectoryContents1.Create();

            string directoryPath2 = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath2);
            IWritableDirectoryContents writableDirectoryContents2 = GetWritableDirectoryContents(directoryPath2);
            writableDirectoryContents2.Create();

            string destinationDirectoryPath = FormattableString.Invariant($"{directoryPath2}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(destinationDirectoryPath);
            writableDirectoryContents1.MoveTo(destinationDirectoryPath);
        }

        [ExpectedException(typeof(DirectoryNotFoundException))]
        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_MoveTo_CompoundPath_NonExistent_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.Create();
            string destinationDirectoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory--{Random.Next()}");
            AssertUtilities.WriteLine(destinationDirectoryPath);
            writableDirectoryContents.MoveTo(destinationDirectoryPath);
        }

        [ExpectedException(typeof(ArgumentException))]
        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_MoveTo_Empty_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.MoveTo("");
        }

        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_MoveTo_Nonexistent_Whitespace_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            try
            {
                writableDirectoryContents.MoveTo("  ");
            }
            catch (IOException ioException)
            {
                AssertUtilities.WriteLine(ioException);
                return;
            }
            Assert.Fail();
        }

        [ExpectedException(typeof(ArgumentNullException))]
        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_MoveTo_Null_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            // ReSharper disable once AssignNullToNotNullAttribute
            writableDirectoryContents.MoveTo(null);
        }

        [ExpectedException(typeof(DirectoryNotFoundException))]
        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_MoveTo_SimplePath_Absolute_NonExistent_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            string destinationDirectoryPath = FormattableString.Invariant($"/TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(destinationDirectoryPath);
            writableDirectoryContents.MoveTo(destinationDirectoryPath);
        }

        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_MoveTo_SimplePath_Absolute_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.Create();
            string destinationDirectoryPath = FormattableString.Invariant($"/TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(destinationDirectoryPath);
            writableDirectoryContents.MoveTo(destinationDirectoryPath);
        }

        [ExpectedException(typeof(DirectoryNotFoundException))]
        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_MoveTo_SimplePath_RelativeInside_NonExistent_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            string destinationDirectoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(destinationDirectoryPath);
            writableDirectoryContents.MoveTo(destinationDirectoryPath);
        }

        [ExpectedException(typeof(IOException))]
        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_MoveTo_SimplePath_RelativeInside_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.Create();
            string destinationDirectoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(destinationDirectoryPath);
            writableDirectoryContents.CreateSubdirectory(destinationDirectoryPath);
            try
            {
                writableDirectoryContents.MoveTo(destinationDirectoryPath);
            }
            catch (IOException ioException)
            {
                AssertUtilities.WriteLine(ioException);
                throw;
            }
        }

        [ExpectedException(typeof(IOException))]
        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_MoveTo_Whitespace_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.Create();
            try
            {
                writableDirectoryContents.MoveTo("  ");
            }
            catch (Exception exception)
            {
                AssertUtilities.WriteLine(exception);
                throw;
            }
        }

        [TestMethod]
        public void IWritableDirectoryContents_SimplePath_WithAFile_Enumeration_Test() =>
            IWritableDirectoryContents_SimplePath_WithAFile_Enumeration_Test(path => GetWritableDirectoryContents(path));

        [TestMethod]
        public void IWritableDirectoryContents_ToString_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            string s = writableDirectoryContents.ToString();
            Assert.IsNotNull(s);
            AssertUtilities.WriteLine(s);
            Assert.IsFalse(string.IsNullOrWhiteSpace(s));
        }

        // IWritableFileInfo Tests

        [TestMethod]
        public void IWritableFileInfo_CompoundPath_Name_Test()
        {
            string directoryName = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            string fileName = FormattableString.Invariant($"TestFile-{Random.Next()}");
            string filePath = FormattableString.Invariant($"{directoryName}\\{fileName}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = GetWritableFileInfo(filePath);
            string name = writableFileInfo.Name;
            AssertUtilities.WriteLine($"Name: {name}");
            Assert.IsNotNull(name);
            Assert.AreEqual(fileName, name);
        }

        [TestMethod]
        public void IWritableFileInfo_CompoundPath_PhysicalPath_Test()
        {
            string directoryName = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            string fileName = FormattableString.Invariant($"TestFile-{Random.Next()}");
            string filePath = FormattableString.Invariant($"{directoryName}\\{fileName}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = GetWritableFileInfo(filePath);
            string physicalPath = writableFileInfo.PhysicalPath;
            AssertUtilities.WriteLine($"PhysicalPath: {physicalPath}");
            if (IsPhysical)
            {
                Assert.IsNotNull(physicalPath);
                // ReSharper disable once AssignNullToNotNullAttribute
                Assert.AreEqual(Path.Combine(PhysicalRoot, filePath), physicalPath);
            }
            else
                Assert.IsNull(physicalPath);
        }

        [TestMethod]
        public void IWritableFileInfo_CreateWriteStream_Write_0B_Test() =>
            IWritableFileInfo_CreateWriteStream_Write_Test(0);

        [TestMethod]
        public void IWritableFileInfo_CreateWriteStream_Write_1B_Test() =>
            IWritableFileInfo_CreateWriteStream_Write_Test(1);

        [TestMethod]
        public void IWritableFileInfo_CreateWriteStream_Write_16B_Test() =>
            IWritableFileInfo_CreateWriteStream_Write_Test(16);

        [TestMethod]
        public void IWritableFileInfo_CreateWriteStream_Write_256B_Test() =>
            IWritableFileInfo_CreateWriteStream_Write_Test(256);

        [TestMethod]
        public void IWritableFileInfo_CreateWriteStream_Write_4KiB_Test() =>
            IWritableFileInfo_CreateWriteStream_Write_Test(4*1024);

        [TestMethod]
        public void IWritableFileInfo_CreateWriteStream_Write_8KiB_Test() =>
            IWritableFileInfo_CreateWriteStream_Write_Test(8 * 1024);

        [TestMethod]
        public void IWritableFileInfo_CreateWriteStream_Write_16KiB_Test() =>
            IWritableFileInfo_CreateWriteStream_Write_Test(16 * 1024);

        [TestMethod]
        public void IWritableFileInfo_CreateWriteStream_Write_32KiB_Test() =>
            IWritableFileInfo_CreateWriteStream_Write_Test(32 * 1024);

        [TestMethod]
        public void IWritableFileInfo_CreateWriteStream_Write_64KiB_Test() => 
            IWritableFileInfo_CreateWriteStream_Write_Test(64 * 1024);

        [TestMethod]
        public void IWritableFileInfo_CreateWriteStream_Write_128KiB_Test() =>
            IWritableFileInfo_CreateWriteStream_Write_Test(128 * 1024);

        [TestMethod]
        public void IWritableFileInfo_CreateWriteStream_SimplePath_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            CreateAndDisposeOfWriteStream(filePath);
        }

        [TestMethod]
        public void IWritableFileInfo_CreateWriteStream_SimplePath_Then_Exists_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = CreateAndDisposeOfWriteStream(filePath);
            Assert.IsTrue(writableFileInfo.Exists);
        }

        [TestMethod]
        public void IWritableFileInfo_Delete_SimplePath_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = CreateAndDisposeOfWriteStream(filePath);
            Assert.IsTrue(writableFileInfo.Exists);
            writableFileInfo.Delete();
        }

        [ExpectedException(typeof(UnauthorizedAccessException))]
        [TestMethod]
        public void IWritableFileInfo_Delete_SimplePath_DirectoryTest()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);
            writableDirectoryContents.Create();
            IWritableFileInfo writableFileInfo = GetWritableFileInfo(directoryPath);
            writableFileInfo.Delete();
        }

        [TestMethod]
        public void IWritableFileInfo_Delete_SimplePath_Failure_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = GetWritableFileInfo(filePath);
            // "If the file does not exist, this method does nothing."
            writableFileInfo.Delete();  // must not throw
        }

        [TestMethod]
        public void IWritableFileInfo_Delete_SimplePath_Then_DoesNotExist_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = CreateAndDisposeOfWriteStream(filePath);
            Assert.IsTrue(writableFileInfo.Exists);
            writableFileInfo.Delete();
            writableFileInfo.Refresh();
            Assert.IsFalse(writableFileInfo.Exists);
        }

        [TestMethod]
        public void IWritableFileInfo_Exists_Failure_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = GetWritableFileInfo(filePath, failing: true);
            bool exists = writableFileInfo.Exists;
            Assert.IsFalse(exists);
        }

        [TestMethod]
        public void IWritableFileInfo_Exists_Negative_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = GetWritableFileInfo(filePath);
            bool exists = writableFileInfo.Exists;
            Assert.IsFalse(exists);
        }

        [TestMethod]
        public void IWritableFileInfo_IsDirectory_Failure_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo fileInfo = GetWritableFileInfo(filePath, failing:true);
            bool exists = fileInfo.IsDirectory;
            Assert.IsFalse(exists);
        }

        [TestMethod]
        public void IWritableFileInfo_IsDirectory_NonExistent_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo fileInfo = GetWritableFileInfo(filePath);
            bool exists = fileInfo.IsDirectory;
            Assert.IsFalse(exists);
        }

        [TestMethod]
        public void IWritableFileInfo_LastModified_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = CreateAndDisposeOfWriteStream(filePath);
            DateTimeOffset lastModified = writableFileInfo.LastModified;
            AssertUtilities.WriteLine(lastModified);
        }

        [TestMethod]
        public void IWritableFileInfo_LastModified_Failure_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = GetWritableFileInfo(filePath, failing: true);
            DateTimeOffset lastModified = writableFileInfo.LastModified;
            AssertUtilities.WriteLine(lastModified);
        }

        [TestMethod]
        public void IWritableFileInfo_LastModified_NonExistent_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = GetWritableFileInfo(filePath);
            DateTimeOffset lastModified = writableFileInfo.LastModified;
            AssertUtilities.WriteLine(lastModified);
        }

        [TestMethod]
        public void IWritableFileInfo_Length_DeletedFile_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = CreateAndDisposeOfWriteStream(filePath);
            writableFileInfo.Delete();
            long length;
            if (IsPhysical)
            {
                try
                {
                    length = writableFileInfo.Length;
                }
                catch (IOException)
                {
                    Assert.Inconclusive("Microsoft's bug in PhysicalFileProvider; we do not want to fail the test.");
                    return;
                }
            }
            else
                length = writableFileInfo.Length;
            Assert.AreEqual(-1, length);
        }

        [TestMethod]
        public void IWritableFileInfo_Length_EmptyFile_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            CreateAndDisposeOfWriteStream(filePath);
            IFileInfo fileInfo = GetFileInfo(filePath);
            Assert.AreEqual(0, fileInfo.Length);
        }

        [ExpectedException(typeof(ArgumentNullException))]
        [TestMethod]
        public void IWritableFileInfo_Name_Null_Test()
        {
            // ReSharper disable once AssignNullToNotNullAttribute
            GetWritableFileInfo(null);
        }

        [TestMethod]
        public void IWritableFileInfo_Replace_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = GetWritableFileInfo(filePath);

            const string testText = "test text";
            writableFileInfo.WriteAllText(testText);

            string destinationFilePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(destinationFilePath);

            writableFileInfo.Replace(destinationFilePath);

            var destinationFileInfo = GetFileInfo(destinationFilePath);

            string textReadBack = destinationFileInfo.ReadAllText();
            Assert.AreEqual(testText, textReadBack);
        }

        [TestMethod]
        public void IWritableFileInfo_SimplePath_Name_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = GetWritableFileInfo(filePath);
            string name = writableFileInfo.Name;
            AssertUtilities.WriteLine($"Name: {name}");
            Assert.IsNotNull(name);
            Assert.AreEqual(filePath, name);
        }

        [TestMethod]
        public void IWritableFileInfo_SimplePath_PhysicalPath_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = GetWritableFileInfo(filePath);
            string physicalPath = writableFileInfo.PhysicalPath;
            AssertUtilities.WriteLine($"PhysicalPath: {physicalPath}");
            if (IsPhysical)
            {
                Assert.IsNotNull(physicalPath);
                // ReSharper disable once AssignNullToNotNullAttribute
                Assert.AreEqual(Path.Combine(PhysicalRoot, filePath), physicalPath);
            }
            else
                Assert.IsNull(physicalPath);
        }

        [TestMethod]
        public void IWritableFileInfo_ToString_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = GetWritableFileInfo(filePath);
            string s = writableFileInfo.ToString();
            Assert.IsNotNull(s);
            AssertUtilities.WriteLine(s);
            Assert.IsFalse(string.IsNullOrWhiteSpace(s));
        }

        [TestMethod]
        public void IWritableFileInfo_WriteAllText_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = GetWritableFileInfo(filePath);

            const string testText = "test text";
            writableFileInfo.WriteAllText(testText);

            string textReadBack = writableFileInfo.ReadAllText();
            Assert.AreEqual(testText, textReadBack);
        }

        // IWritableFileProvider

        [TestMethod]
        public void IWritableFileProvider_GetWritableDirectoryContents_Empty_Test()
        {
            GetWritableDirectoryContents("");
        }

        [TestMethod]
        public void IWritableFileProvider_GetWritableDirectoryContents_Null_Test()
        {
            GetWritableDirectoryContents(null);
        }

        [TestMethod]
        public void IWritableFileProvider_GetWritableDirectoryContents_CompoundPath_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            GetWritableDirectoryContents(directoryPath);
        }

        [TestMethod]
        public void IWritableFileProvider_GetWritableDirectoryContents_SimplePath_Test()
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            GetWritableDirectoryContents(directoryPath);
        }

        [TestMethod]
        public void IWritableFileProvider_GetWritableFileInfo_CompoundPath_Test()
        {
            string filePath = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            GetWritableFileInfo(filePath);
        }

        [TestMethod]
        public void IWritableFileProvider_GetWritableFileInfo_SimplePath_Test()
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            GetWritableFileInfo(filePath);
        }

        [ExpectedException(typeof(ArgumentNullException))]
        [TestMethod]
        public void IWritableFileProvider_GetWritableFileInfo_Null_Test()
        {
            // ReSharper disable once AssignNullToNotNullAttribute
            GetWritableFileInfo(null);
        }

        [TestMethod]
        public void IWritableFileProvider_ToString_Test()
        {
            IWritableFileProvider writableFileProvider = CreateWritableFileProvider(failing: false);
            string s = writableFileProvider.ToString();
            Assert.IsNotNull(s);
            AssertUtilities.WriteLine(s);
            Assert.IsFalse(string.IsNullOrWhiteSpace(s));
        }

        #endregion



        #region Protected Properties

        protected bool IsPhysical => !(PhysicalRoot is null);

        protected abstract long MaximumFileLength { get; }

        [CanBeNull]
        protected abstract string PhysicalRoot { get; }

        #endregion




        #region Protected Methods

        [NotNull]
        protected abstract IFileProvider CreateFileProvider(bool failing);

        [NotNull]
        protected abstract IFileProvider2 CreateFileProvider2(bool failing);

        [NotNull]
        protected abstract IWritableFileProvider CreateWritableFileProvider(bool failing);

        #endregion




        #region Private Methods

        [NotNull]
        private IWritableFileInfo CreateAndDisposeOfWriteStream([NotNull] string filePath)
        {
            IWritableFileInfo writableFileInfo = GetWritableFileInfo(filePath);
            using (Stream stream = writableFileInfo.CreateWriteStream())
            {
                Assert.IsNotNull(stream);
                Assert.IsTrue(stream.CanWrite);
                if (stream.CanSeek)
                {
                    Assert.AreEqual(0, stream.Length);
                    Assert.AreEqual(0, stream.Position);
                }
            }
            return writableFileInfo;
        }

        [NotNull]
        private IDirectoryContents EnumerateDirectoryContents(
            [NotNull] Func<string, IDirectoryContents> getDirectoryContentsFunction,
            [CanBeNull] string directoryPath)
        {
            IDirectoryContents directoryContents = getDirectoryContentsFunction(directoryPath);
            Check.AssertNotNull(directoryContents);
            foreach (IFileInfo fileInfo in directoryContents)
            {
                Assert.IsNotNull(fileInfo);
                Assert.IsNotNull(fileInfo.Name);
                AssertUtilities.WriteLine($"Name: {fileInfo.Name}");
                //if (!fileInfo.Exists)
                //    ;
                //if (fileInfo.IsDirectory)
                //    ;
                Assert.IsTrue(fileInfo.Exists);
                if (IsPhysical)
                    Assert.IsNotNull(fileInfo.PhysicalPath);
                else
                    Assert.IsNull(fileInfo.PhysicalPath);
            }
            return directoryContents;
        }

        [NotNull]
        private IWritableDirectoryContents EnumerateWritableDirectoryContents(
            [NotNull] Func<string, IWritableDirectoryContents> getWritableDirectoryContentsFunction,
            [CanBeNull] string directoryPath)
        {
            IWritableDirectoryContents writableDirectoryContents = getWritableDirectoryContentsFunction(directoryPath);
            Check.AssertNotNull(writableDirectoryContents);
            foreach (IFileInfo fileInfo in writableDirectoryContents)
            {
                Assert.IsNotNull(fileInfo);
                Assert.IsNotNull(fileInfo.Name);
                AssertUtilities.WriteLine($"Name: {fileInfo.Name}");
                Assert.IsInstanceOfType(fileInfo, typeof(IWritableFileInfo));
                Assert.IsTrue(fileInfo.Exists);
                if (IsPhysical)
                    Assert.IsNotNull(fileInfo.PhysicalPath);
                else
                    Assert.IsNull(fileInfo.PhysicalPath);
            }
            return writableDirectoryContents;
        }

        [NotNull]
        private IDirectoryContents GetDirectoryContents([CanBeNull] string directoryPath, bool failing = false)
        {
            IFileProvider fileProvider = CreateFileProvider(failing);
            IDirectoryContents directoryContents = fileProvider.GetDirectoryContents(directoryPath);
            Assert.IsNotNull(directoryContents);
            return directoryContents;
        }

        [NotNull]
        private IDirectoryContents2 GetDirectoryContents2([CanBeNull] string directoryPath, bool failing = false)
        {
            IFileProvider2 fileProvider2 = CreateFileProvider2(failing);
            IDirectoryContents2 directoryContents2 = fileProvider2.GetDirectoryContents2(directoryPath);
            Assert.IsNotNull(directoryContents2);
            return directoryContents2;
        }

        [NotNull]
        private IFileInfo GetFileInfo([CanBeNull] string filePath, bool failing = false)
        {
            IFileProvider fileProvider = CreateFileProvider(failing);
            IFileInfo fileInfo = fileProvider.GetFileInfo(filePath);
            Assert.IsNotNull(fileInfo);
            return fileInfo;
        }

        [NotNull]
        private IFileInfo2 GetFileInfo2([CanBeNull] string filePath, bool failing = false)
        {
            IFileProvider2 fileProvider2 = CreateFileProvider2(failing);
            IFileInfo2 fileInfo2 = fileProvider2.GetFileInfo2(filePath);
            Assert.IsNotNull(fileInfo2);
            return fileInfo2;
        }

        [NotNull]
        private IWritableDirectoryContents GetWritableDirectoryContents([CanBeNull] string filePath, bool failing = false)
        {
            IWritableFileProvider writableFileProvider = CreateWritableFileProvider(failing);
            IWritableDirectoryContents writableDirectoryContents = writableFileProvider.GetWritableDirectoryContents(filePath);
            Assert.IsNotNull(writableDirectoryContents);
            return writableDirectoryContents;
        }

        [NotNull]
        private IWritableFileInfo GetWritableFileInfo([CanBeNull] string filePath, bool failing = false)
        {
            IWritableFileProvider writableFileProvider = CreateWritableFileProvider(failing);
            // ReSharper disable once AssignNullToNotNullAttribute
            IWritableFileInfo writableFileInfo = writableFileProvider.GetWritableFileInfo(filePath);
            Assert.IsNotNull(writableFileInfo);
            return writableFileInfo;
        }

        // IDirectoryContents_XXXX

        private void IDirectoryContents_CompoundPath_WithAFile_Enumeration_Test(
            [NotNull] Func<string, IDirectoryContents> getDirectoryContentsFunction)
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine($"{nameof(directoryPath)}: {directoryPath}");
            GetWritableDirectoryContents(directoryPath).Create();

            string fileName = FormattableString.Invariant($"TestFile-{Random.Next()}");
            string filePath = FormattableString.Invariant($"{directoryPath}\\{fileName}");
            AssertUtilities.WriteLine(filePath);
            CreateAndDisposeOfWriteStream(filePath);
            IDirectoryContents directoryContents = EnumerateDirectoryContents(getDirectoryContentsFunction, directoryPath);
            // ReSharper disable once PossibleNullReferenceException
            IFileInfo fileInfo = directoryContents.FirstOrDefault(info => info.Name == fileName);
            Assert.IsNotNull(fileInfo);
            Assert.IsTrue(fileInfo.Exists);
            Assert.IsFalse(fileInfo.IsDirectory);
            if (IsPhysical)
                Assert.IsNotNull(fileInfo.PhysicalPath);
            else
                Assert.IsNull(fileInfo.PhysicalPath);
            Assert.AreEqual(0, fileInfo.Length);
        }

        private void IDirectoryContents_EmptyPath_Enumeration_Test(
            [NotNull] Func<string, IDirectoryContents> getDirectoryContentsFunction) => 
            EnumerateDirectoryContents(getDirectoryContentsFunction, "");

        private void IDirectoryContents_EmptyPath_ToString_Test([NotNull] Func<string, IDirectoryContents> getDirectoryContentsFunction)
        {
            IDirectoryContents directoryContents = getDirectoryContentsFunction("");
            Check.AssertNotNull(directoryContents);
            string s = directoryContents.ToString();
            Assert.IsNotNull(s);
            AssertUtilities.WriteLine(s);
            Assert.IsFalse(string.IsNullOrWhiteSpace(s));
        }

        private void IDirectoryContents_EmptyPath_WithAFile_Enumeration_Test(
            [NotNull] Func<string, IDirectoryContents> getDirectoryContentsFunction)
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            CreateAndDisposeOfWriteStream(filePath);
            IDirectoryContents directoryContents = EnumerateDirectoryContents(getDirectoryContentsFunction, "");
            // ReSharper disable once PossibleNullReferenceException
            IFileInfo fileInfo = directoryContents.FirstOrDefault(info => info.Name == filePath);
            Assert.IsNotNull(fileInfo);
            Assert.IsTrue(fileInfo.Exists);
            Assert.IsFalse(fileInfo.IsDirectory);
            if (IsPhysical)
                Assert.IsNotNull(fileInfo.PhysicalPath);
            else
                Assert.IsNull(fileInfo.PhysicalPath);
            Assert.AreEqual(0, fileInfo.Length);
        }

        private void IDirectoryContents_SimplePath_WithAFile_Enumeration_Test(
            [NotNull] Func<string, IDirectoryContents> getDirectoryContentsFunction)
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine($"{nameof(directoryPath)}: {directoryPath}");
            GetWritableDirectoryContents(directoryPath).Create();

            string fileName = FormattableString.Invariant($"TestFile-{Random.Next()}");
            string filePath = FormattableString.Invariant($"{directoryPath}\\{fileName}");
            AssertUtilities.WriteLine(filePath);
            CreateAndDisposeOfWriteStream(filePath);
            IDirectoryContents directoryContents = EnumerateDirectoryContents(getDirectoryContentsFunction, directoryPath);
            // ReSharper disable once PossibleNullReferenceException
            IFileInfo fileInfo = directoryContents.FirstOrDefault(info => info.Name == fileName);
            Assert.IsNotNull(fileInfo);
            Assert.IsTrue(fileInfo.Exists);
            Assert.IsFalse(fileInfo.IsDirectory);
            if (IsPhysical)
                Assert.IsNotNull(fileInfo.PhysicalPath);
            else
                Assert.IsNull(fileInfo.PhysicalPath);
            Assert.AreEqual(0, fileInfo.Length);
        }

        private void IDirectoryContents_GetStatistics_Test(bool recursive)
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine($"{nameof(directoryPath)}: {directoryPath}");
            IWritableDirectoryContents writableDirectoryContents = GetWritableDirectoryContents(directoryPath);

            string filePath1 = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine($"{nameof(filePath1)}: {filePath1}");
            writableDirectoryContents.Create();
            IWritableFileInfo writableFileInfo1 = writableDirectoryContents.GetWritableFileInfo(filePath1);
            writableFileInfo1.WriteAllBytes(new byte[10]);

            string directoryPath2 = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine($"{nameof(directoryPath2)}: {directoryPath2}");
            IWritableDirectoryContents writableDirectoryContents2 = writableDirectoryContents.CreateSubdirectory(directoryPath2);

            string filePath2 = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine($"{nameof(filePath2)}: {filePath2}");
            IWritableFileInfo writableFileInfo2 = writableDirectoryContents2.GetWritableFileInfo(filePath2);
            writableFileInfo2.WriteAllBytes(new byte[5]);

            (long totalBytes, long totalFiles, long totalDirectories) = writableDirectoryContents.GetStatistics(recursive);
            AssertUtilities.WriteLine($"totalBytes: {totalBytes}");
            AssertUtilities.WriteLine($"totalFiles: {totalFiles}");
            AssertUtilities.WriteLine($"totalDirectories: {totalDirectories}");

            Assert.AreEqual(recursive ? 15 : 10, totalBytes, "totalBytes");
            Assert.AreEqual(recursive ? 2 : 1, totalFiles, "totalFiles");
            Assert.AreEqual(1, totalDirectories, "totalDirectories");
        }

        private void IDirectoryContents_NullPath_Enumeration_Test(
            [NotNull] Func<string, IDirectoryContents> getDirectoryContentsFunction) =>
            EnumerateDirectoryContents(getDirectoryContentsFunction, null);

        private void IDirectoryContents_Exists_Negative_Test([NotNull] Func<string, IDirectoryContents> getDirectoryContentsFunction)
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            IDirectoryContents directoryContents = getDirectoryContentsFunction(directoryPath);
            Check.AssertNotNull(directoryContents);
            bool exists = directoryContents.Exists;
            Assert.IsFalse(exists);
        }

        // IFileInfo_XXXX

        private void IFileInfo_CreateReadStream_CompoundPath_Test([NotNull] Func<string, IFileInfo> getFileInfoFunction)
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine(directoryPath);
            GetWritableDirectoryContents(directoryPath).Create();

            string filePath = FormattableString.Invariant($"{directoryPath}\\TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            CreateAndDisposeOfWriteStream(filePath);
            IFileInfo fileInfo = getFileInfoFunction(filePath);
            Check.AssertNotNull(fileInfo);
            using (Stream stream = fileInfo.CreateReadStream())
            {
                Assert.IsNotNull(stream);
                Assert.IsTrue(stream.CanRead);
                if (stream.CanSeek)
                {
                    Assert.AreEqual(0, stream.Length);
                    Assert.AreEqual(0, stream.Position);
                }
            }
        }

        private void IFileInfo_CreateReadStream_NonExistent_Test([NotNull] Func<string, IFileInfo> getFileInfoFunction)
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IFileInfo fileInfo = getFileInfoFunction(filePath);
            Check.AssertNotNull(fileInfo);
            fileInfo.CreateReadStream();
        }

        private void IFileInfo_CreateReadStream_Read_Empty_Test(
            [NotNull] Func<string, IFileInfo> getFileInfoFunction,
            int length)
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = GetWritableFileInfo(filePath);
            using (Stream _ = writableFileInfo.CreateWriteStream())
                // ReSharper disable once EmptyEmbeddedStatement
#pragma warning disable CS0642 // Possible mistaken empty statement
                /*do nothing*/ ;
#pragma warning restore CS0642 // Possible mistaken empty statement

            IFileInfo fileInfo = getFileInfoFunction(filePath);
            Check.AssertNotNull(fileInfo);
            var buffer = new byte[length];
            using (Stream stream = fileInfo.CreateReadStream())
                // ReSharper disable once PossibleNullReferenceException
                stream.Read(buffer, 0, length);
        }

        private void IFileInfo_CreateReadStream_Read_Full_Test(
            [NotNull] Func<string, IFileInfo> getFileInfoFunction,
            int length)
        {
            if (length > MaximumFileLength)
                Assert.Inconclusive();
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = GetWritableFileInfo(filePath);
            const int writeBufferSize = 8192;
            var writeBuffer = new byte[writeBufferSize];
            using (Stream writeStream = writableFileInfo.CreateWriteStream())
            {
                int bytesLeft = length;
                while (bytesLeft != 0)
                {
                    int count = Math.Min(writeBufferSize, bytesLeft);
                    writeStream.Write(writeBuffer, 0, count);
                    bytesLeft -= count;
                }
            }

            IFileInfo fileInfo = getFileInfoFunction(filePath);
            Check.AssertNotNull(fileInfo);
            var readBuffer = new byte[length];
            using (Stream readStream = fileInfo.CreateReadStream())
                // ReSharper disable once PossibleNullReferenceException
                readStream.Read(readBuffer, 0, length);
        }

        private void IFileInfo_CreateReadStream_SimplePath_Test([NotNull] Func<string, IFileInfo> getFileInfoFunction)
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            CreateAndDisposeOfWriteStream(filePath);
            IFileInfo fileInfo = getFileInfoFunction(filePath);
            Check.AssertNotNull(fileInfo);
            using (Stream stream = fileInfo.CreateReadStream())
            {
                Assert.IsNotNull(stream);
                Assert.IsTrue(stream.CanRead);
                if (stream.CanSeek)
                {
                    Assert.AreEqual(0, stream.Length);
                    Assert.AreEqual(0, stream.Position);
                }
            }
        }

        private void IFileInfo_Exists_Negative_Test([NotNull] Func<string, IFileInfo> getFileInfoFunction)
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IFileInfo fileInfo = getFileInfoFunction(filePath);
            Check.AssertNotNull(fileInfo);
            bool exists = fileInfo.Exists;
            Assert.IsFalse(exists);
        }


        private void IFileInfo_IsDirectory_NonExistent_Test([NotNull] Func<string, IFileInfo> getFileInfoFunction)
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IFileInfo fileInfo = getFileInfoFunction(filePath);
            Check.AssertNotNull(fileInfo);
            bool exists = fileInfo.IsDirectory;
            Assert.IsFalse(exists);
        }

        private void IFileInfo_Length_DeletedFile_Test([NotNull] Func<string, IFileInfo> getFileInfoFunction)
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = CreateAndDisposeOfWriteStream(filePath);
            IFileInfo fileInfo = getFileInfoFunction(filePath);
            Check.AssertNotNull(fileInfo);
            writableFileInfo.Delete();
            long length = fileInfo.Length;
            AssertUtilities.WriteLine(length);
            // The IFileInfo and IFileInfo2 may or may not be caching, and if it is caching, it may or may not invalidate
            // itself on IWritableFileInfo.Delete. The proper Length can thus either be 0 (reflecting the original length),
            // or -1 (reflecting the fact that the file has been deleted).
            Assert.IsTrue((length == 0) || (length == -1));
        }

        private void IFileInfo_Length_NonExistent_Test([NotNull] Func<string, IFileInfo> getFileInfoFunction)
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IFileInfo fileInfo = getFileInfoFunction(filePath);
            Check.AssertNotNull(fileInfo);
            long length;
            if (IsPhysical)
            {
                try
                {
                    length = fileInfo.Length;
                }
                catch (IOException)
                {
                    Assert.Inconclusive("Microsoft's bug in PhysicalFileProvider; we do not want to fail the test.");
                    return;
                }
            }
            else
                length = fileInfo.Length;
            Assert.AreEqual(-1, length);
        }

        private void IFileInfo_Name_Empty_Test([NotNull] Func<string, IFileInfo> getFileInfoFunction)
        {
            IFileInfo fileInfo = getFileInfoFunction("");
            Check.AssertNotNull(fileInfo);
            string name = fileInfo.Name;
            AssertUtilities.WriteLine($"Name: {name}");
            Assert.AreEqual("", name);
        }

        private void IFileInfo_Name_Null_Test([NotNull] Func<string, IFileInfo> getFileInfoFunction)
        {
            IFileInfo fileInfo = getFileInfoFunction(null);
            Check.AssertNotNull(fileInfo);
            string name = fileInfo.Name;
            AssertUtilities.WriteLine($"Name: {name}");
            Assert.IsNull(name);
        }

        // IFileProvider_XXXX

        private void IFileProvider_ToString_Test([NotNull] IFileProvider fileProvider)
        {
            string s = fileProvider.ToString();
            Assert.IsNotNull(s);
            AssertUtilities.WriteLine(s);
            Assert.IsFalse(string.IsNullOrWhiteSpace(s));
        }

        private void IFileProvider_Watch_Null_Test([NotNull] IFileProvider fileProvider)
        {
            IChangeToken changeToken = fileProvider.Watch(null);
            Assert.IsNotNull(changeToken);
        }

        // IWritableDirectoryContents

        private void IWritableDirectoryContents_CompoundPath_WithAFile_Enumeration_Test(
            [NotNull] Func<string, IWritableDirectoryContents> getWritableDirectoryContentsFunction)
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}\\TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine($"{nameof(directoryPath)}: {directoryPath}");
            GetWritableDirectoryContents(directoryPath).Create();

            string fileName = FormattableString.Invariant($"TestFile-{Random.Next()}");
            string filePath = FormattableString.Invariant($"{directoryPath}\\{fileName}");
            AssertUtilities.WriteLine(filePath);
            CreateAndDisposeOfWriteStream(filePath);
            IWritableDirectoryContents directoryContents = EnumerateWritableDirectoryContents(getWritableDirectoryContentsFunction, directoryPath);
            // ReSharper disable once PossibleNullReferenceException
            IFileInfo fileInfo = directoryContents.FirstOrDefault(info => info.Name == fileName);
            Assert.IsNotNull(fileInfo);
            Assert.IsTrue(fileInfo.Exists);
            Assert.IsFalse(fileInfo.IsDirectory);
            if (IsPhysical)
                Assert.IsNotNull(fileInfo.PhysicalPath);
            else
                Assert.IsNull(fileInfo.PhysicalPath);
            Assert.AreEqual(0, fileInfo.Length);
        }

        private void IWritableDirectoryContents_EmptyPath_Enumeration_Test(
            [NotNull] Func<string, IWritableDirectoryContents> getWritableDirectoryContentsFunction) =>
            EnumerateWritableDirectoryContents(getWritableDirectoryContentsFunction, "");

        private void IWritableDirectoryContents_EmptyPath_WithAFile_Enumeration_Test(
            [NotNull] Func<string, IWritableDirectoryContents> getWritableDirectoryContentsFunction)
        {
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            CreateAndDisposeOfWriteStream(filePath);
            IWritableDirectoryContents writableDirectoryContents = EnumerateWritableDirectoryContents(getWritableDirectoryContentsFunction, "");
            // ReSharper disable once PossibleNullReferenceException
            IFileInfo fileInfo = writableDirectoryContents.FirstOrDefault(info => info.Name == filePath);
            Assert.IsNotNull(fileInfo);
            Assert.IsTrue(fileInfo.Exists);
            Assert.IsFalse(fileInfo.IsDirectory);
            if (IsPhysical)
                Assert.IsNotNull(fileInfo.PhysicalPath);
            else
                Assert.IsNull(fileInfo.PhysicalPath);
            Assert.AreEqual(0, fileInfo.Length);
        }

        private void IWritableDirectoryContents_NullPath_Enumeration_Test(
            [NotNull] Func<string, IWritableDirectoryContents> getWritableDirectoryContentsFunction) =>
            EnumerateWritableDirectoryContents(getWritableDirectoryContentsFunction, null);

        private void IWritableDirectoryContents_SimplePath_WithAFile_Enumeration_Test(
            [NotNull] Func<string, IWritableDirectoryContents> getWritableDirectoryContentsFunction)
        {
            string directoryPath = FormattableString.Invariant($"TestDirectory-{Random.Next()}");
            AssertUtilities.WriteLine($"{nameof(directoryPath)}: {directoryPath}");
            GetWritableDirectoryContents(directoryPath).Create();

            string fileName = FormattableString.Invariant($"TestFile-{Random.Next()}");
            string filePath = FormattableString.Invariant($"{directoryPath}\\{fileName}");
            AssertUtilities.WriteLine(filePath);
            CreateAndDisposeOfWriteStream(filePath);
            IWritableDirectoryContents writableDirectoryContents = EnumerateWritableDirectoryContents(getWritableDirectoryContentsFunction, directoryPath);
            // ReSharper disable once PossibleNullReferenceException
            IFileInfo fileInfo = writableDirectoryContents.FirstOrDefault(info => info.Name == fileName);
            Assert.IsNotNull(fileInfo);
            Assert.IsTrue(fileInfo.Exists);
            Assert.IsFalse(fileInfo.IsDirectory);
            if (IsPhysical)
                Assert.IsNotNull(fileInfo.PhysicalPath);
            else
                Assert.IsNull(fileInfo.PhysicalPath);
            Assert.AreEqual(0, fileInfo.Length);
        }

        // IWritableFileInfo

        private void IWritableFileInfo_CreateWriteStream_Write_Test(int length)
        {
            if (length > MaximumFileLength)
                Assert.Inconclusive();
            string filePath = FormattableString.Invariant($"TestFile-{Random.Next()}");
            AssertUtilities.WriteLine(filePath);
            IWritableFileInfo writableFileInfo = GetWritableFileInfo(filePath);
            var writeBuffer = new byte[length];
            using (Stream writeStream = writableFileInfo.CreateWriteStream())
                writeStream.Write(writeBuffer, 0, length);
        }

        #endregion
    }
}
