﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib;
using OpcLabs.UnitTestLib;

namespace OpcLabs.BaseLibTest
{
    public partial class ResourceDescriptorTest
    {
        [TestMethod]
        public void CtorTest_String_Empty()
        {
            var target = new ResourceDescriptor("");
            Assert.AreEqual("", target.ToString());
        }

        [TestMethod]
        public void CtorTest_String_NoScheme_NoAuthority_Path1()
        {
            var target = new ResourceDescriptor("testSegment");
            Assert.AreEqual("testSegment", target.ToString());
        }

        [TestMethod]
        public void CtorTest_String_NoScheme_Authority_NoPath()
        {
            var target = new ResourceDescriptor("//testAuthority");
            Assert.AreEqual("//testAuthority", target.ToString());
        }

        [TestMethod]
        public void CtorTest_String_NoScheme_Authority_Path1()
        {
            var target = new ResourceDescriptor("//testAuthority/testSegment");
            Assert.AreEqual("//testAuthority/testSegment", target.ToString());
        }

        [TestMethod]
        public void IsNullTest_Negative()
        {
            var target = new ResourceDescriptor("testUrlString");
            Assert.IsFalse(target.IsNull);
        }

        [TestMethod]
        public void IsNullTest_Positive()
        {
            var target = ResourceDescriptor.Null;
            Assert.IsTrue(target.IsNull);
        }
        
        [TestMethod]
        public void ToStringTest_Null()
        {
            var target = ResourceDescriptor.Null;
            Assert.AreEqual("", target.ToString());
        }

        [TestMethod]
        public void EqualityTest_NewObjects()
        {
            var target1 = new ResourceDescriptor();
            var target2 = new ResourceDescriptor();
            Assert.AreEqual(target1.GetHashCode(), target2.GetHashCode());
            Assert.AreEqual(target1, target2);
        }

        [TestMethod]
        public void EqualityTest_Copy()
        {
            var target1 = new ResourceDescriptor("http://172.22.144.53/XmlDaSampleServer/Service.asmx");
            AssertUtilities.WriteLine(target1);
            var target2 = new ResourceDescriptor(target1);
            AssertUtilities.WriteLine(target2);
            Assert.AreEqual(target1.GetHashCode(), target2.GetHashCode());
            Assert.AreEqual(target1, target2);
        }
    }
}
