﻿using System;
using System.Threading;
using JetBrains.Annotations;
using OpcLabs.BaseLib.ComInterop;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.AddressSpace;
using OpcLabs.EasyOpcLiveTest.DataAccess.Environment;

namespace OpcLabs.EasyOpcLiveTest.DataAccess
{
    static class KitServerCatalogEntry
    {
        [NotNull]
        static public DAServerCatalogEntry Create([NotNull] string location)
        {
            // ReSharper disable RedundantArgumentName

            var catalogEntry = new DAServerCatalogEntry(new ServerElement(
                location:location, 
                clsid:Guid.Parse(OpcLabs_KitServer.ClsidString), 
                progId:OpcLabs_KitServer.ProgId, 
                versionIndependentProgId: OpcLabs_KitServer.VersionIndependentProgId,
                description:"OPC Labs Kit Server",
                vendor:"OPC Labs, http://www.opclabs.com",
                serverCategories:new ServerCategories {OpcDataAccess20 = true, OpcDataAccess30 = true},
                scheme:ServerDescriptor.OpcDAScheme));

            // / Simulation/Constant_*

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name:"Constant_BOOL", 
                browsePath:"/Simulation/Constant_BOOL", 
                nodePath:null, 
                itemId:"Simulation.Constant_BOOL", 
                isBranch:false, 
                isLeaf:true, 
                hasChildren:false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Readable, 
                BehaviorKind = OpcItemBehaviorKind.Constant,
                DataType = VarTypes.Bool, 
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = false,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                // ReSharper disable StringLiteralTypo
                name: "Constant_BSTR",
                browsePath: "/Simulation/Constant_BSTR",
                nodePath: null,
                itemId: "Simulation.Constant_BSTR",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
                // ReSharper restore StringLiteralTypo
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Readable,
                BehaviorKind = OpcItemBehaviorKind.Constant,
                DataType = VarTypes.BStr,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = "",
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Constant_CY",
                browsePath: "/Simulation/Constant_CY",
                nodePath: null,
                itemId: "Simulation.Constant_CY",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Readable,
                BehaviorKind = OpcItemBehaviorKind.Constant,
                DataType = VarTypes.Cy,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = (Decimal)0,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Constant_DATE",
                browsePath: "/Simulation/Constant_DATE",
                nodePath: null,
                itemId: "Simulation.Constant_DATE",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Readable,
                BehaviorKind = OpcItemBehaviorKind.Constant,
                DataType = VarTypes.Date,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = new DateTime(1899, 12, 30),
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Constant_DECIMAL",
                browsePath: "/Simulation/Constant_DECIMAL",
                nodePath: null,
                itemId: "Simulation.Constant_DECIMAL",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Readable,
                BehaviorKind = OpcItemBehaviorKind.Constant,
                DataType = VarTypes.Decimal,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = (Decimal)0,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Constant_EMPTY",
                browsePath: "/Simulation/Constant_EMPTY",
                nodePath: null,
                itemId: "Simulation.Constant_EMPTY",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Readable,
                BehaviorKind = OpcItemBehaviorKind.Constant,
                DataType = VarTypes.Empty,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = null,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Constant_I1",
                browsePath: "/Simulation/Constant_I1",
                nodePath: null,
                itemId: "Simulation.Constant_I1",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Readable,
                BehaviorKind = OpcItemBehaviorKind.Constant,
                DataType = VarTypes.I1,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = (short)0,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Constant_I2",
                browsePath: "/Simulation/Constant_I2",
                nodePath: null,
                itemId: "Simulation.Constant_I2",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Readable,
                BehaviorKind = OpcItemBehaviorKind.Constant,
                DataType = VarTypes.I2,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = (short)0,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Constant_I4",
                browsePath: "/Simulation/Constant_I4",
                nodePath: null,
                itemId: "Simulation.Constant_I4",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Readable,
                BehaviorKind = OpcItemBehaviorKind.Constant,
                DataType = VarTypes.I4,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = 0,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Constant_R4",
                browsePath: "/Simulation/Constant_R4",
                nodePath: null,
                itemId: "Simulation.Constant_R4",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Readable,
                BehaviorKind = OpcItemBehaviorKind.Constant,
                DataType = VarTypes.R4,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = 0.0f,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Constant_R8",
                browsePath: "/Simulation/Constant_R8",
                nodePath: null,
                itemId: "Simulation.Constant_R8",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Readable,
                BehaviorKind = OpcItemBehaviorKind.Constant,
                DataType = VarTypes.R8,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = 0.0d,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Constant_UI1",
                browsePath: "/Simulation/Constant_UI1",
                nodePath: null,
                itemId: "Simulation.Constant_UI1",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Readable,
                BehaviorKind = OpcItemBehaviorKind.Constant,
                DataType = VarTypes.UI1,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = (byte)0,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Constant_UI2",
                browsePath: "/Simulation/Constant_UI2",
                nodePath: null,
                itemId: "Simulation.Constant_UI2",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Readable,
                BehaviorKind = OpcItemBehaviorKind.Constant,
                DataType = VarTypes.UI2,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = 0,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Constant_UI4",
                browsePath: "/Simulation/Constant_UI4",
                nodePath: null,
                itemId: "Simulation.Constant_UI4",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Readable,
                BehaviorKind = OpcItemBehaviorKind.Constant,
                DataType = VarTypes.UI4,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = (long)0,
            });

            // / Simulation/Register_*

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Register_BOOL",
                browsePath: "/Simulation/Register_BOOL",
                nodePath: null,
                itemId: "Simulation.Register_BOOL",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Both,
                BehaviorKind = OpcItemBehaviorKind.Register,
                DataType = VarTypes.Bool,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = false,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                // ReSharper disable StringLiteralTypo
                name: "Register_BSTR",
                browsePath: "/Simulation/Register_BSTR",
                nodePath: null,
                itemId: "Simulation.Register_BSTR",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
                // ReSharper restore StringLiteralTypo
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Both,
                BehaviorKind = OpcItemBehaviorKind.Register,
                DataType = VarTypes.BStr,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = "",
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Register_CY",
                browsePath: "/Simulation/Register_CY",
                nodePath: null,
                itemId: "Simulation.Register_CY",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Both,
                BehaviorKind = OpcItemBehaviorKind.Register,
                DataType = VarTypes.Cy,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = (Decimal)0,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Register_DATE",
                browsePath: "/Simulation/Register_DATE",
                nodePath: null,
                itemId: "Simulation.Register_DATE",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Both,
                BehaviorKind = OpcItemBehaviorKind.Register,
                DataType = VarTypes.Date,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = new DateTime(1899, 12, 30),
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Register_DECIMAL",
                browsePath: "/Simulation/Register_DECIMAL",
                nodePath: null,
                itemId: "Simulation.Register_DECIMAL",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Both,
                BehaviorKind = OpcItemBehaviorKind.Register,
                DataType = VarTypes.Decimal,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = (Decimal)0,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Register_EMPTY",
                browsePath: "/Simulation/Register_EMPTY",
                nodePath: null,
                itemId: "Simulation.Register_EMPTY",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Both,
                BehaviorKind = OpcItemBehaviorKind.Register,
                DataType = VarTypes.Empty,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = null,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Register_I1",
                browsePath: "/Simulation/Register_I1",
                nodePath: null,
                itemId: "Simulation.Register_I1",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Both,
                BehaviorKind = OpcItemBehaviorKind.Register,
                DataType = VarTypes.I1,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = (short)0,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Register_I2",
                browsePath: "/Simulation/Register_I2",
                nodePath: null,
                itemId: "Simulation.Register_I2",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Both,
                BehaviorKind = OpcItemBehaviorKind.Register,
                DataType = VarTypes.I2,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = (short)0,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Register_I4",
                browsePath: "/Simulation/Register_I4",
                nodePath: null,
                itemId: "Simulation.Register_I4",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Both,
                BehaviorKind = OpcItemBehaviorKind.Register,
                DataType = VarTypes.I4,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = 0,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Register_R4",
                browsePath: "/Simulation/Register_R4",
                nodePath: null,
                itemId: "Simulation.Register_R4",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Both,
                BehaviorKind = OpcItemBehaviorKind.Register,
                DataType = VarTypes.R4,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = 0.0f,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Register_R8",
                browsePath: "/Simulation/Register_R8",
                nodePath: null,
                itemId: "Simulation.Register_R8",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Both,
                BehaviorKind = OpcItemBehaviorKind.Register,
                DataType = VarTypes.R8,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = 0.0d,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Register_UI1",
                browsePath: "/Simulation/Register_UI1",
                nodePath: null,
                itemId: "Simulation.Register_UI1",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Both,
                BehaviorKind = OpcItemBehaviorKind.Register,
                DataType = VarTypes.UI1,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = (byte)0,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Register_UI2",
                browsePath: "/Simulation/Register_UI2",
                nodePath: null,
                itemId: "Simulation.Register_UI2",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Both,
                BehaviorKind = OpcItemBehaviorKind.Register,
                DataType = VarTypes.UI2,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = 0,
            });

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Register_UI4",
                browsePath: "/Simulation/Register_UI4",
                nodePath: null,
                itemId: "Simulation.Register_UI4",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Both,
                BehaviorKind = OpcItemBehaviorKind.Register,
                DataType = VarTypes.UI4,
                FastestRateOfChange = Timeout.Infinite,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = Timeout.Infinite,
                Value = (long)0,
            });

            // ...

            catalogEntry.ItemCatalog.Add(new DAItemCatalogEntry(new DANodeElement(
                name: "Random (1 s)",
                browsePath: "/Simulation/Random (1 s)",
                nodePath: null,
                itemId: "Simulation.Random (1 s)",
                isBranch: false,
                isLeaf: true,
                hasChildren: false))
            {
                PropertyCatalog = MakeCommonPropertyCatalog(),
                AccessRights = DAAccessRights.Readable,
                BehaviorKind = OpcItemBehaviorKind.Unknown,
                DataType = VarTypes.R8,
                FastestRateOfChange = 0,
                Quality = DAQualities.GoodNonspecific,
                SlowestRateOfChange = 0,
            });
            


            // TODO: need some write-only item
            // TODO: many others...

            return catalogEntry;
            // ReSharper restore RedundantArgumentName
        }

        [NotNull]
        static private DAPropertyCatalog MakeCommonPropertyCatalog()
        {
            var propertyCatalog = new DAPropertyCatalog
            {
                new DAPropertyCatalogEntry(new DAPropertyElement(DAPropertyIds.DataType, VarTypes.I2, "Item Canonical Data Type")),
                new DAPropertyCatalogEntry(new DAPropertyElement(DAPropertyIds.Value, VarTypes.Empty, "Item Value")),
                new DAPropertyCatalogEntry(new DAPropertyElement(DAPropertyIds.Quality, VarTypes.I2, "Item Quality")),
                new DAPropertyCatalogEntry(new DAPropertyElement(DAPropertyIds.Timestamp, VarTypes.Date, "Item Timestamp")),
                new DAPropertyCatalogEntry(new DAPropertyElement(DAPropertyIds.AccessRights, VarTypes.I4, "Item Access Rights")),
                new DAPropertyCatalogEntry(new DAPropertyElement(DAPropertyIds.ScanRate, VarTypes.R4, "Server Scan Rate")),
                new DAPropertyCatalogEntry(new DAPropertyElement(DAPropertyIds.EUType, VarTypes.I4, "Item EU Type")),

                //new DAPropertyCatalogEntry(new DAPropertyElement(DAPropertyIds.HighEU, VarTypes.R8, "High EU")),
                //new DAPropertyCatalogEntry(new DAPropertyElement(DAPropertyIds.LowEU, VarTypes.R8, "Low EU")),

                // TODO: have more
            };

            return propertyCatalog;
        }
    }
}
