﻿using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.DataAccess.AddressSpace;
using OpcLabs.EasyOpcFactory.DataAccess.AddressSpace;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcTest.DataAccess.AddressSpace
{
    public partial class DAPropertyElementCollectionTest
    {
        [TestMethod]
        public void LookupTest()
        {
            DAPropertyElement targetElement0 = DAPropertyElementInstances.Instance0;
            DAPropertyElement targetElement1 = DAPropertyElementInstances.Instance1;
            var target = new DAPropertyElementCollection {targetElement0, targetElement1};

            Assert.IsTrue(target.Contains(targetElement0.PropertyId));
            Assert.IsTrue(target.Contains(targetElement1.PropertyId));

            DAPropertyElement actualElement0 = target[targetElement0.PropertyId];
            Assert.AreEqual(targetElement0, actualElement0);

            DAPropertyElement actualElement1 = target[targetElement1.PropertyId];
            Assert.AreEqual(targetElement1, actualElement1);
        }

        [TestCategory("BinarySerialization")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BinarySerializationTest0()
        {
            Tests.BinarySerializationTest(CreateSerializationTarget0());
        }

        [TestCategory("BinarySerialization")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BinarySerializationTest1()
        {
            Tests.BinarySerializationTest(CreateSerializationTarget1());
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTest0()
        {
            Tests.XmlSerializationTest(CreateSerializationTarget0());
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTest1()
        {
            Tests.XmlSerializationTest(CreateSerializationTarget1());
        }

        [NotNull]
        private DAPropertyElementCollection CreateSerializationTarget0()
        {
            return new DAPropertyElementCollection();
        }

        [NotNull]
        private DAPropertyElementCollection CreateSerializationTarget1()
        {
            var propertyElementCollection = new DAPropertyElementCollection
                                            {
                                                DAPropertyElementInstances.Instance0,
                                                DAPropertyElementInstances.Instance1,
                                                DAPropertyElementInstances.Instance2,
                                                DAPropertyElementInstances.Instance3,
                                                DAPropertyElementInstances.Instance4
                                            };
            return propertyElementCollection;
        }
    }
}
