﻿using System;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcTest.DataAccess
{
    public partial class DAVtqTest
    {
        #region Static Private Methods

        [NotNull]
        static private DAVtq CreateSerializationTarget0()
        {
            return DAVtq.Null;
        }

        [NotNull]
        static private DAVtq CreateSerializationTarget1()
        {
            return new DAVtq(null, DateTime.MinValue, DAQualities.GoodNonspecific);
        }

        [NotNull]
        static private DAVtq CreateSerializationTarget2()
        {
            return new DAVtq(-546372819, DateTime.MinValue, DAQualities.GoodNonspecific);
        }

        [NotNull]
        static private DAVtq CreateSerializationTarget3()
        {
            return new DAVtq("TestValue", DateTime.MinValue, DAQualities.GoodNonspecific);
        }

        #endregion



        #region Test Methods

        [TestCategory("BinarySerialization")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BinarySerializationTest0()
        {
            Tests.BinarySerializationTest(CreateSerializationTarget0());
        }

        [TestCategory("BinarySerialization")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BinarySerializationTest1()
        {
            Tests.BinarySerializationTest(CreateSerializationTarget1());
        }

        [TestCategory("BinarySerialization")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BinarySerializationTest2()
        {
            Tests.BinarySerializationTest(CreateSerializationTarget2());
        }

        [TestCategory("BinarySerialization")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BinarySerializationTest3()
        {
            Tests.BinarySerializationTest(CreateSerializationTarget3());
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void DisplayValueTest0()
        {
            var target = new DAVtq();
            Assert.AreEqual("null", target.DisplayValue());
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void DisplayValueTest1()
        {
            var target = new DAVtq(value:1);
            Assert.AreEqual("1", target.DisplayValue());
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTest0()
        {
            var target = new DAVtq();
            Assert.AreEqual("null {} @1601-01-01T00:00:00.000; GoodNonspecific (192)", target.ToString());
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTest1()
        {
            // ReSharper disable once RedundantArgumentName
            var target = new DAVtq(value: null, quality:DAQualities.BadNonspecific);
            Assert.AreEqual("@1601-01-01T00:00:00.000; BadNonspecific LimitOk (0)", target.ToString());
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTest2()
        {
            var target = new DAVtq(value: 1);
            Assert.AreEqual("1 {Int32} @1601-01-01T00:00:00.000; GoodNonspecific (192)", target.ToString());
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTest3()
        {
            // ReSharper disable once RedundantArgumentName
            var target = new DAVtq(value: 1, timestamp: new DateTime(1999, 11, 28, 22, 58, 58, DateTimeKind.Utc));
            Assert.AreEqual("1 {Int32} @1999-11-28T22:58:58.000; GoodNonspecific (192)", target.ToString());
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTest0()
        {
            Tests.XmlSerializationTest(CreateSerializationTarget0());
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTest1()
        {
            Tests.XmlSerializationTest(CreateSerializationTarget1());
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTest2()
        {
            Tests.XmlSerializationTest(CreateSerializationTarget2());
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTest3()
        {
            Tests.XmlSerializationTest(CreateSerializationTarget3());
        }

        #endregion
    }
}
