﻿using System;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.ComInterop;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcTest.DataAccess.OperationModel
{
    public partial class DAItemVtqArgumentsTest
    {
        [TestCategory("BinarySerialization")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BinarySerializationTest0()
        {
            Tests.BinarySerializationTest(CreateSerializationTarget0());
        }

        [TestCategory("BinarySerialization")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BinarySerializationTest1()
        {
            Tests.BinarySerializationTest(CreateSerializationTarget1());
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTest0()
        {
            Tests.XmlSerializationTest(CreateSerializationTarget0());
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTest1()
        {
            Tests.XmlSerializationTest(CreateSerializationTarget1());
        }

        [TestMethod]
        public void CtorTest_String_String_String_VarType_Object_DateTime_DAQuality_Object()
        {
            var dateTime = new DateTime(2014, 1, 1, 0, 0, 0, DateTimeKind.Utc);

            var target = new DAItemVtqArguments(
                "TestMachineName", "TestServerClass", 
                "TestItemId", VarTypes.R8,
                111.22, dateTime, DAQualities.GoodNonspecific, 
                "TestState");
            Assert.AreEqual(new ServerDescriptor("TestMachineName", "TestServerClass"), target.ServerDescriptor);
            Assert.AreEqual(new DAItemDescriptor("TestItemId", VarTypes.R8), target.ItemDescriptor);
            Assert.AreEqual(new DAVtq(111.22, dateTime, DAQualities.GoodNonspecific), target.Vtq);
            Assert.AreEqual("TestState", target.State);
        }

        [NotNull]
        private DAItemVtqArguments CreateSerializationTarget0()
        {
            return new DAItemVtqArguments();
        }

        [NotNull]
        private DAItemVtqArguments CreateSerializationTarget1()
        {
            return new DAItemVtqArguments("machine1", "company1.class1",
                "branch1.item1", VarTypes.I4, "TestValue", DateTime.Now.ToUniversalTime(), DAQualities.ConfigError, state: 88888888);
        }
    }
}
